//
// Asynchronous Read-only cache 
//
// Design by G.J. van Loo, FenLogic Ltd, November-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
//
// Altough this is a fully finctiopnal and synthesiable design,
// it is not very realistic module as a-synchronous memories are no 
// longer used in ASIC's.(They can still be made from registers 
// or FPGA LUT's but are not area effecient.)
// It serves to show principle of how a cache works.
//
// It can also be used to show the increase in complexity 
// resulting from the obligatory use of synchronous memories by
// comparting it to a normal (synchronous) read-only cache
// 
// Addtional required modules:
//   async_tag.v
//   async_read_mem.v
//


module async_ro_cache 
#(parameter 
   DATAWIDTH = 32, // Width of data
   ADRSWIDTH = 16, // Width of address
   L2DEPTH   =  8  // log-2 size of the cache 
  )
(
   // System
   input   clk,
   input   reset_n,
   
   // Control
   input       flush,
    
   // Read port
   input                  read,
   input  [ADRSWIDTH-1:0] radrs,
   output                 valid,
   output [DATAWIDTH-1:0] rdata,
   
   // External (main) memory port
   output                 mem_read,
   output [ADRSWIDTH-1:0] mem_radrs,
   input                  mem_data_valid,
   input  [DATAWIDTH-1:0] mem_rdata
   
   );

localparam TAG_WIDTH = L2DEPTH;  
localparam MATCH_WIDTH =  ADRSWIDTH - TAG_WIDTH; 


wire   [TAG_WIDTH-1:0] tag_adrs ;
wire [MATCH_WIDTH-1:0] tag_data ;
wire flushing;

wire                 cache_mem_write;
wire [DATAWIDTH-1:0] cache_mem_rdata ;


wire hit;

   // split incoming address up in tag (and cache memory) address
   // and tag match section
   assign tag_adrs = radrs[TAG_WIDTH-1:0];
   assign tag_data = radrs[ADRSWIDTH-1:TAG_WIDTH];
   
   // external memory read on a miss
   assign mem_read = !flushing & read & !hit;

   // may get synthysis warning as this is a pass-through
   assign mem_radrs = radrs; 
   
   // Output is valid on a hit or if data arrives from main memory 
   // This assumes mem_data_valid is retracted after data has arrived. 
   assign valid = hit | mem_data_valid;
   
   // Data comes from cache on hit, outside on miss
   assign rdata = hit ? cache_mem_rdata : mem_rdata;
   
   // Write cache when data comes from main memory 
   // This assumes mem_data_valid is retracted after data has arrived. 
   assign cache_mem_write = mem_data_valid;
   
   
async_tag
   #(
      .TAG_WIDTH   (TAG_WIDTH)  ,
      .MATCH_WIDTH (MATCH_WIDTH)
   ) // parameters
async_tag_0 (

   // System
      .clk      (clk),
      .reset_n  (reset_n),
   
   // Control
      .flush    (flush),
      .flushing (flushing),
   
   // Tag lookup
      .tag_adrs (tag_adrs),
      .tag_data (tag_data),
   
   // Lookup result
      .hit      (hit),
   // Update tag mem 
      .tag_update(mem_data_valid)
   );
   
   
// Asynchronous read memory for tags and valid bits 
async_read_mem
  #(
      .WIDTH  (DATAWIDTH), // width in bits
      .L2DEPTH(TAG_WIDTH)  // Log 2 Depth 
   ) // parameters
tag_mem (
      // write port 
      .clk  (clk),             // write clock       
               
      .write(cache_mem_write), // write access 
      .wadrs(tag_adrs),        // write address
      .wdata(mem_rdata),       // write data
   
      // read port 
      .radrs(tag_adrs),  // read address
      .rdata(cache_mem_rdata) 
   );
   





   
endmodule
   
   
