//
// Asynchronous tag handling 
// Part of the asynchronous cache module 
// (See comments in header there)
// 
// Design by G.J. van Loo, FenLogic Ltd, November-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
//
// This module handles the cache tag.
// it decides if there is a cache hit or miss.
// In case of a miss the tag memory is updated. 
//
// Addtional required modules:
//   async_read_mem.v
// 

module async_tag
#(parameter
    TAG_WIDTH  = 8,  // == Number of cache entries 
    MATCH_WIDTH=12   // Address bits we match aginst
  )
(
   // System
   input   clk,
   input   reset_n,
   
   // Control
   input       flush,
   output reg  flushing,
   
   // Tag lookup
   input    [TAG_WIDTH-1:0] tag_adrs,
   input  [MATCH_WIDTH-1:0] tag_data,
   
   // Match result: one cycle later
   output  reg    hit,
   
   // Tell when data is arriving so we know when we can continue
   input          tag_update
   
   );

// Need address counter for flushing
reg    [TAG_WIDTH-1:0] flush_count;

// When writing memory select between flush & update
wire   [TAG_WIDTH-1:0] tagmem_wadrs;
wire [MATCH_WIDTH  :0] tagmem_wdata;
wire                   tagmem_write;

// Tags read from tag memory   
wire [MATCH_WIDTH  :0] tagmem_rdata;
wire [MATCH_WIDTH-1:0] mem_match;
wire                   mem_valid;

   assign tagmem_write = flushing ? 1'b1 : tag_update;
   assign tagmem_wadrs = flushing ? flush_count : tag_adrs;
   assign tagmem_wdata = flushing ? {(MATCH_WIDTH+1){1'b0}} : {1'b1,tag_data};

// Asynchronous read memory for tags and valid bits 
async_read_mem
  #(
      .WIDTH  (MATCH_WIDTH+1), // width in bits
      .L2DEPTH(TAG_WIDTH)      // Log 2 Depth 
   ) // parameters
tag_mem (
      // write port 
      .clk  (clk),             // write clock       
               
      .write(tagmem_write),       // write access 
      .wadrs(tagmem_wadrs),       // write address
      .wdata(tagmem_wdata),       // write data
   
      // read port 
      .radrs(tag_adrs),  // read address
      .rdata(tagmem_rdata) 
   );
   
   // split read data in match & valid bit
   assign mem_match = tagmem_rdata[MATCH_WIDTH-1:0]; 
   assign mem_valid = tagmem_rdata[MATCH_WIDTH]; 
  
   // Here we do the actual match
   // Whilst flushing we cannot have a hit
   always @( * )
   begin
      if (!flushing && mem_valid && tag_data==mem_match)
         hit = 1'b1; 
      else
         hit = 1'b0;
   end
   
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         flushing    <= 1'b1;
         flush_count <= {TAG_WIDTH{1'b0}};
      end
      else      
      begin
         // Flush control
         // Multiple flush requests do not start a new flush when
         // the previous one is still in progress.
         // This also means the 'flush' input can be toggled slowly
         // e.g. by a CPU write-set followed by a write-clear
         if (flush & !flushing)
            flushing <= 1'b1;
            // flush counter is zero here
         else
            if (flush_count=={TAG_WIDTH{1'b1}})
               flushing <= 1'b0;
         
         if (flushing)
            flush_count <= flush_count + 1;
                          
      end
   end
   
endmodule
   
   
