//
// Read-only two-way cache 
//
// Design by G.J. van Loo, FenLogic Ltd, November-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
// 
// This is the two-way version of the read-only cache. 
// (Which is a one-way cache)
// 
// Addtional required modules:
//   tag2.v
//   spm.v (Synchronous singleport memory)
//
//


module ro_cache2 
#(parameter 
   DATAWIDTH = 32, // Width of data
   ADRSWIDTH = 16, // Width of address
   L2DEPTH   =  8  // Log 2 number of cache data entries 
  )
(
   // System
   input   clk,
   input   reset_n,
   
   // Control
   input       flush,
    
   // Read port
   input                  read,
   input  [ADRSWIDTH-1:0] radrs,
   output                 ready,
   output [DATAWIDTH-1:0] rdata,
   
   // External (main) memory port
   output                 mem_read,
   output [ADRSWIDTH-1:0] mem_radrs,
   input                  mem_data_valid,
   input  [DATAWIDTH-1:0] mem_rdata
   
   );

// For a two-way cache the memories are half as deep
// (But there are two of them) 
localparam L2MEMDEPTH  = L2DEPTH-1;

wire [L2MEMDEPTH-1:0] cmem_rd_adrs,cmem_wt_adrs ;
wire                  hit0_1,hit1_1;
wire                  valid_1;
wire                  bank_select;
wire                  write_bank0;
wire                  write_bank1;
wire                  flushing;

// Because of synchronous memory we need the previous
// address in case there is a miss. 
// We also need this when wrting the cache
reg  [ADRSWIDTH-1:0] radrs_1;

// If we write the cache or tag we can not at the same time read
// Thus we can NOT give a ready until AFTER the write.
// But by then the data from the main memory has gone
// So we have to hold it for one cycle longer
reg [DATAWIDTH-1:0] mem_rdata_1;

// Switch cache memory adress between read & write 
wire  [L2MEMDEPTH-1:0] cache_mem_adrs;

wire                 cache_mem_write;
wire [DATAWIDTH-1:0] cache_mem_rdata0_1 ;
wire [DATAWIDTH-1:0] cache_mem_rdata1_1 ;
wire [DATAWIDTH-1:0] cache_mem_rdata_1 ;

   // Get memory adress from incoming address 
   assign cmem_rd_adrs = radrs[L2MEMDEPTH-1:0];
   // Same for when we write to tag using delayed address
   assign cmem_wt_adrs = radrs_1[L2MEMDEPTH-1:0];
   
   // external memory read on a miss
   assign mem_read = valid_1 & !hit0_1 & !hit1_1;

   // main mem read address is prev. address
   assign mem_radrs = radrs_1; 
      
   // Data comes from cache on hit, outside on miss
   // Need to use hit for which one to use 
   assign cache_mem_rdata_1 = hit0_1 ? cache_mem_rdata0_1 : cache_mem_rdata1_1; 
   assign rdata = valid_1 & (hit0_1 | hit1_1)? cache_mem_rdata_1 : mem_rdata_1;
   
   // Write cache when data comes from main memory 
   // This assumes mem_data_valid is retracted after data has arrived. 
   assign cache_mem_write = mem_data_valid;  
   
   // select read or write address
   assign cache_mem_adrs = mem_data_valid ? cmem_wt_adrs : cmem_rd_adrs;
   
tag2
   #(
      .L2DEPTH   (L2DEPTH)  ,
      .ADRSWIDTH (ADRSWIDTH)
   ) // parameters
tag2_0 (

   // System
      .clk      (clk),
      .reset_n  (reset_n),
   
   // Control
      .flush    (flush),
      .flushing (flushing),

   // Tag lookup
      .adrs_valid(read),
      .radrs     (radrs),
   
   // Lookup result
      .valid_1  (valid_1),
      .hit0_1   (hit0_1),
      .hit1_1   (hit1_1),
      .ready    (ready),
      
   // Update tag mem 
      .bank_select(bank_select),
      .tag_update (mem_data_valid),
      .wt_adrs    (radrs_1)
   );
   
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         // Stricly speaking these do not need a reset
         radrs_1     <= 'b0;
         mem_rdata_1 <= 'b0;
      end
      else      
      begin
         if (ready)
            radrs_1  <= radrs;
         if (mem_data_valid)
            mem_rdata_1 <= mem_rdata;
         else
            // If we have a cache hit and flushing we would lose the data
            // So here we catch it for emitting after the flush
            if (valid_1 & (hit0_1 | hit1_1) & flushing)
               mem_rdata_1 <= cache_mem_rdata_1;
      end
   end
  

// Main memory of the cache: the user data
// (twice)
spm
  #(
      .WIDTH  (DATAWIDTH), // width in bits
      .L2DEPTH(L2MEMDEPTH)    // Log 2 Depth 
   ) // parameters
cache_mem0 (
      // write port 
      .clk     (clk),              // System clock   
      .enable  (1'b1),             // No power saving yet
               
      .write   (write_bank0),      // write  
      .address (cache_mem_adrs),   // read/write adddress
      .wdata   (mem_rdata),        // write data
      .rdata   (cache_mem_rdata0_1) // read data ( 1 cycle delay)
   );

spm
  #(
      .WIDTH  (DATAWIDTH), // width in bits
      .L2DEPTH(L2MEMDEPTH)    // Log 2 Depth 
   ) // parameters
cache_mem1 (
      // write port 
      .clk     (clk),              // System clock   
      .enable  (1'b1),             // No power saving yet
               
      .write   (write_bank1),      // write  
      .address (cache_mem_adrs),   // read/write adddress
      .wdata   (mem_rdata),        // write data
      .rdata   (cache_mem_rdata1_1) // read data ( 1 cycle delay)
   );
   
   assign write_bank0 = bank_select==1'b0 ? mem_data_valid : 1'b0;
   assign write_bank1 = bank_select==1'b1 ? mem_data_valid : 1'b0;
   
   
endmodule
   
   
