//
// Read-only cache 
// Single bank (1-way) synchronous single port memory
// This module uses the fast-flush TAG version (ff_tag.v)
//
// Design by G.J. van Loo, FenLogic Ltd, December-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
// 
// This is the follow up step from the asynchronous cache
// It uses one synchronous single ported memories
// This is a more realistic cache which can be used in designs
//
// Addtional required modules:
//   ff_tag.v
//   spm.v (Synchronous singleport memory)
//
//


module ro_ff_cache 
#(parameter 
   DATAWIDTH    = 32, // Width of data
   ADRSWIDTH    = 16, // Width of address
   L2DEPTH      =  8, // Log 2 number of cache data entries 
   L2VALIDWIDTH = 4   // Log 2 valid memory data width 
  )
(
   // System
   input   clk,
   input   reset_n,
   
   // Control
   input       flush,
    
   // Read port
   input                  read,
   input  [ADRSWIDTH-1:0] radrs,
   output                 ready,
   output [DATAWIDTH-1:0] rdata,
   
   // External (main) memory port
   output                 mem_read,
   output [ADRSWIDTH-1:0] mem_radrs,
   input                  mem_data_valid,
   input  [DATAWIDTH-1:0] mem_rdata
   
   );

// For a one-way cache these are identical   
localparam L2MEMDEPTH  = L2DEPTH;


wire  [L2MEMDEPTH-1:0] cmem_rd_adrs,cmem_wt_adrs ;
wire                   hit_1;
wire                   valid_1;
wire                   flushing;

// Because of synchronous memory we need the previous
// address in case there is a miss. 
// We also need this when wrting the cache
reg    [ADRSWIDTH-1:0] radrs_1;

// If we write the cache or tag we can not at the same time read
// Thus we can NOT give a ready until AFTER the write.
// But by then the data from the main memory has gone
// So we have to hold it for one cycle longer
reg [DATAWIDTH-1:0] mem_rdata_1;

// Switch cache memory adress between read & write 
wire  [L2MEMDEPTH-1:0] cache_mem_adrs;

wire                 cache_mem_write;
wire [DATAWIDTH-1:0] cache_mem_rdata_1 ;

   // Get memory adress from incoming address 
   
   assign cmem_rd_adrs = radrs[L2MEMDEPTH-1:0];
   // Same for when we write to tag using delayed address
   assign cmem_wt_adrs = radrs_1[L2MEMDEPTH-1:0];
   
   // external memory read on a miss
   assign mem_read = valid_1 & !hit_1;

   // main mem read address is prev. address
   assign mem_radrs = radrs_1; 
      
   // Data comes from cache on hit, outside on miss
   assign rdata = valid_1 & hit_1 ? cache_mem_rdata_1 : mem_rdata_1;
   
   // Write cache when data comes from main memory 
   // This assumes mem_data_valid is retracted after data has arrived. 
   assign cache_mem_write = mem_data_valid;  
   
   // select read or write address
   assign cache_mem_adrs = mem_data_valid ? cmem_wt_adrs : cmem_rd_adrs;
   
ff_tag
   #(
      .L2DEPTH     (L2DEPTH)  ,
      .ADRSWIDTH   (ADRSWIDTH),
      .L2VALIDWIDTH(L2VALIDWIDTH)
   ) // parameters
ff_tag_0 (

   // System
      .clk      (clk),
      .reset_n  (reset_n),
   
   // Control
      .flush    (flush),
      .flushing (flushing),
   
   // Tag lookup
      .adrs_valid(read),
      .radrs     (radrs),
   
   // Lookup result
      .valid_1  (valid_1),
      .hit_1    (hit_1),
      .ready    (ready),
      
   // Update tag mem 
      .tag_update(mem_data_valid),
      .wt_adrs   (radrs_1)
   );
   
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         // Stricly speaking these do not need a reset
         radrs_1     <= 'b0;
         mem_rdata_1 <= 'b0;
      end
      else      
      begin
         if (ready)
            radrs_1  <= radrs;
         if (mem_data_valid)
            mem_rdata_1 <= mem_rdata;
         else
            // If we have a cache hit and flushing we would lose the data
            // So here we catch it for emitting after the flush
            if (valid_1 & hit_1 & flushing)
               mem_rdata_1 <= cache_mem_rdata_1;
               
      end
   end
  

// Main memory of the cache: the user data
spm
  #(
      .WIDTH  (DATAWIDTH), // width in bits
      .L2DEPTH(L2MEMDEPTH)    // Log 2 Depth 
   ) // parameters
cache_mem (
      // write port 
      .clk     (clk),              // System clock   
      .enable  (1'b1),             // No power saving yet
               
      .write   (mem_data_valid),   // write access 
      .address (cache_mem_adrs),   // read/write adddress
      .wdata   (mem_rdata),        // write data
      .rdata   (cache_mem_rdata_1) // read data ( 1 cycle delay)
   );
   
endmodule
   
   
