//
// Read-only line cache 
// Single bank (1-way) synchronous single port memory
// 4 data units lines
//
// Design by G.J. van Loo, FenLogic Ltd, November-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
// 
// This is a variant of the read only cache (ro_cache.v) but it
// operates on line reads (a series of reads) instead of singletons
// If you check the code you will see that it it uses the standard 
// tag.v from the ro_cache.v design.
// But it uses different parameters giving a smaller tag unit. 
// 
// Addtional required modules:
//   tag.v
//   spm.v (Synchronous singleport memory)
//

module ro_linecache 
#(parameter 
   DATAWIDTH = 32, // Width of data
   ADRSWIDTH = 16, // Width of address
   L2DEPTH   =  8, // Log 2 number of cache data entries 
   L2LINE    =  2  // Log 2 line length 
  )
(
   // System
   input   clk,
   input   reset_n,
   
   // Control
   input       flush,
    
   // Read port
   input                  read,
   input  [ADRSWIDTH-1:0] radrs,
   output                 ready,
   output [DATAWIDTH-1:0] rdata,
   
   // External (main) memory port
   // Beware that for a request we expect a burts of data back!
   output                 mem_read,
   output [ADRSWIDTH-1:0] mem_radrs,
   input                  mem_data_valid,
   input  [DATAWIDTH-1:0] mem_rdata
   
   );

// Data DEPTH same for one-way line cache 
localparam L2MEMDEPTH  = L2DEPTH;
localparam LINELENGTH  = 1<<L2LINE;

// The TAG admin does not care about the bits which address the line
localparam TAGADRBITS = ADRSWIDTH-L2LINE;

wire     [TAGADRBITS-1:0] tag_radrs,tag_wadrs;
wire     [L2MEMDEPTH-1:0] cmem_rd_adrs,cmem_wt_adrs ;
wire         [L2LINE-1:0] line_radrs;
reg          [L2LINE-1:0] line_radrs_1; // which unit in a line 
wire                      hit_1;
wire                      valid_1;
wire                      flushing;

// Because of synchronous memory we need the previous
// address in case there is a miss. 
// We also need this when wrting the cache
reg    [ADRSWIDTH-1:0] radrs_1;

// If we write the cache or tag we can not at the same time read
// Thus we can NOT give a ready until AFTER the write.
// But by then the data from the main memory has gone
// So we have to hold it for one cycle longer
// We use this also to hold the right element from a line when it arrives
reg [DATAWIDTH-1:0] mem_rdata_1;

// Switch cache memory adress between read & write 
wire  [L2MEMDEPTH-1:0] cache_mem_adrs;

wire                 cache_mem_write; 
wire [DATAWIDTH-1:0] cache_mem_rdata_1 ;

// When burst arrives from main mmeory need a counter
reg     [L2LINE-1:0] wrt_count; 
wire    [L2LINE-1:0] line_wadrs; // which unit in a line 

   // Tag does not need to know about lines
   assign tag_radrs = radrs  [ADRSWIDTH-1:L2LINE];
   assign tag_wadrs = radrs_1[ADRSWIDTH-1:L2LINE];
   
   // Which unit of the line we need:
   assign line_radrs = radrs[L2LINE-1:0];
   // Same for when we write to tag using delayed address
   assign line_wadrs = radrs_1[L2LINE-1:0]; 
 
   // Get cache memory adress from incoming address 
   assign cmem_rd_adrs = radrs[L2MEMDEPTH-1:0];
   
   // We write a line to memory for that we use
   // THE MS address bits and a write counter
   assign cmem_wt_adrs = {radrs_1[L2MEMDEPTH-1:L2LINE],wrt_count};
   
   // external memory read on a miss
   assign mem_read = valid_1 & !hit_1;

   // main mem read address is MSBits of prev. address
   // LS adress bits are zero (but can also be dropped)
   assign mem_radrs = {radrs_1[ADRSWIDTH-1:L2LINE],{L2LINE{1'b0}}}; 
      
   // Data comes from cache on hit, outside on miss
   assign rdata = valid_1 & hit_1 ? cache_mem_rdata_1 : mem_rdata_1;
   
   // Write cache when data comes from main memory 
   // This assumes mem_data_valid is retracted after data has arrived. 
   assign cache_mem_write = mem_data_valid;  
   
   // select read or write address
   assign cache_mem_adrs = mem_data_valid ? cmem_wt_adrs : cmem_rd_adrs;
   
// Tag memory only tags cache lines
// So it can be smaller   
// For the rest it functions identical
tag
   #(
      .L2DEPTH   (L2DEPTH-L2LINE),
      .ADRSWIDTH (ADRSWIDTH-L2LINE)
   ) // parameters
tag_0 (

   // System
      .clk      (clk),
      .reset_n  (reset_n),
   
   // Control
      .flush    (flush),
      .flushing (flushing),
   
   // Tag lookup
      .adrs_valid(read),
      .radrs     (tag_radrs),
   
   // Lookup result
      .valid_1  (valid_1),
      .hit_1    (hit_1),
      .ready    (ready),
      
   // Update tag mem when last data arrives
      .tag_update(mem_data_valid && (wrt_count==LINELENGTH-1)),
      .wt_adrs   (tag_wadrs)
   );
   
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         wrt_count   <= {L2LINE{1'b0}};
         // Stricly speaking these do not need a reset
         radrs_1     <= 'b0;
         mem_rdata_1 <= 'b0;
         line_radrs_1<= {L2LINE{1'b0}};
      end
      else      
      begin
         if (ready)
         begin
            radrs_1  <= radrs;
            line_radrs_1 <= line_radrs;
         end
         
         if (mem_data_valid)
         begin
            wrt_count <= wrt_count+1;
            // When data arrives pick the right unit from the line 
            // To later pass on to the user 
            if (wrt_count==line_radrs_1)
               mem_rdata_1 <= mem_rdata;
         end
         else
            // If we have a cache hit and flushing we would lose the data
            // So here we catch it for emitting after the flush
            if (valid_1 & hit_1 & flushing)
               mem_rdata_1 <= cache_mem_rdata_1;
      end
   end
  

// Main memory of the cache: the user data
spm
  #(
      .WIDTH  (DATAWIDTH), // width in bits
      .L2DEPTH(L2MEMDEPTH)    // Log 2 Depth 
   ) // parameters
cache_mem (
      // write port 
      .clk     (clk),              // System clock   
      .enable  (1'b1),             // No power saving yet
               
      .write   (mem_data_valid),   // write access 
      .address (cache_mem_adrs),   // read/write adddress
      .wdata   (mem_rdata),        // write data
      .rdata   (cache_mem_rdata_1) // read data ( 1 cycle delay)
   );
      
endmodule
   
   
