//
// Synchronous two-way tag module 
// 
// Design by G.J. van Loo, FenLogic Ltd, December-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
//
// This module is a variant of the one-way tag (tag.v) 
// 
// Addtional required modules:
//   spm.v (Synchronous singleport memory)
   
module tag2
#(parameter
    L2DEPTH  =12,  // Log 2 number of cache data entries 
    ADRSWIDTH=30   // Address bits neede to access cache data units
  )
(
   // System
   input   clk,
   input   reset_n,
   
   // Control
   input       flush,
   output reg  flushing,
   
   // Tag lookup
   input                  adrs_valid,
   input  [ADRSWIDTH-1:0] radrs,
   
   // Match result: one cycle later
   output  reg    valid_1,
   output  reg    hit0_1,
   output  reg    hit1_1,
   // indicate when we are ready  
   output         ready,
   
   // Tag update 
   output reg            bank_select,
   input                 tag_update,
   input [ADRSWIDTH-1:0] wt_adrs
   
   );

// For a two-way cache the memories are half as deep
// (But there are two of them) 
localparam L2MEMDEPTH  = L2DEPTH-1;

// We need to match all address bits which are not used
// to address the tag memory.
localparam MATCH_WIDTH = ADRSWIDTH - L2MEMDEPTH;
 
// Need address counter plus flag for flushing
reg    [L2MEMDEPTH-1:0] flush_count;
reg                     flushing_1;

// When writing memory select between flush & update
wire [MATCH_WIDTH  :0] tagmem_wdata;
wire                   tagmem_write;
// Which of the two memories we write 
wire                   tagmem_write0;
wire                   tagmem_write1;

// Read from tag memory   
reg   [L2MEMDEPTH-1:0] tagmem_adrs;
wire [MATCH_WIDTH  :0] tagmem_rdata0_1;
wire [MATCH_WIDTH  :0] tagmem_rdata1_1;
wire [MATCH_WIDTH-1:0] mem_match0_1;
wire [MATCH_WIDTH-1:0] mem_match1_1;
wire                   mem_valid0_1;
wire                   mem_valid1_1;

// Read address for tag memory
wire [L2MEMDEPTH-1:0]  tag_adrs;

// Match-against data now and one cycle later
wire [MATCH_WIDTH-1:0] tag;
reg  [MATCH_WIDTH-1:0] tag_1;

// Write address split into adressa and tag
wire  [L2MEMDEPTH-1:0] write_adrs;
wire [MATCH_WIDTH-1:0] write_tag;

// Pseudo random 'way' select when writing
reg  [1:0] mem_valid;
reg        pseudo_random;


// The most difficult part: data flow control
// To do with ready & wait_for_write & flushing
reg                    wait_for_write;

   // Split the address in a tag and match
   assign tag_adrs   = radrs[L2MEMDEPTH-1:0];
   assign tag        = radrs[ADRSWIDTH-1:L2MEMDEPTH];
   assign write_adrs = wt_adrs[L2MEMDEPTH-1:0];
   assign write_tag  = wt_adrs[ADRSWIDTH-1:L2MEMDEPTH];
   
   // not ready on miss (both!), waiting or flushing 
   assign ready = ! ((valid_1 & !hit0_1 & !hit1_1) | wait_for_write | flushing);

   // Write to tag memory when flusing or updating
   // All other times we read (Thus no power saving!)
   assign tagmem_write = flushing ? 1'b1 : tag_update;
   
   // Clear valid bits when flushing
   // Might as well clear the whole memory
   assign tagmem_wdata = flushing ? {(MATCH_WIDTH+1){1'b0}} : {1'b1,write_tag};
   
   always @( * )
   begin
      if (flushing)
         tagmem_adrs = flush_count;
      else
         if (tag_update)
            tagmem_adrs = write_adrs;
         else
            tagmem_adrs = tag_adrs;
   end


// Synchronous read memory for tags and valid bits 
// (Twice)
spm
  #(
      .WIDTH  (MATCH_WIDTH+1), // width in bits
      .L2DEPTH(L2MEMDEPTH)      // Log 2 Depth 
   ) // parameters
tag_mem0 (
      // write port 
      .clk     (clk),             // write clock  
      .enable  (1'b1),     
               
      .write   (tagmem_write0),  // write access 
      .address (tagmem_adrs),    // read/write address
      .wdata   (tagmem_wdata),   // write data
      .rdata   (tagmem_rdata0_1) // read data ( 1 cycle delay)
   );

spm
  #(
      .WIDTH  (MATCH_WIDTH+1), // width in bits
      .L2DEPTH(L2MEMDEPTH)      // Log 2 Depth 
   ) // parameters
tag_mem1 (
      // write port 
      .clk     (clk),             // write clock  
      .enable  (1'b1),     
               
      .write   (tagmem_write1),  // write access 
      .address (tagmem_adrs),    // read/write address
      .wdata   (tagmem_wdata),   // write data
      .rdata   (tagmem_rdata1_1) // read data ( 1 cycle delay)
   );
    
   // split read data in match & valid bit
   assign mem_match0_1 = tagmem_rdata0_1[MATCH_WIDTH-1:0]; 
   assign mem_valid0_1 = tagmem_rdata0_1[MATCH_WIDTH]; 
   assign mem_match1_1 = tagmem_rdata1_1[MATCH_WIDTH-1:0]; 
   assign mem_valid1_1 = tagmem_rdata1_1[MATCH_WIDTH]; 
  
   // Here we do the actual match
   // Whilst flushing we cannot have a hit
   always @( * )
   begin
      if (!flushing_1 && mem_valid0_1 && tag_1==mem_match0_1)
         hit0_1 = 1'b1; 
      else
         hit0_1 = 1'b0;
      if (!flushing_1 && mem_valid1_1 && tag_1==mem_match1_1)
         hit1_1 = 1'b1; 
      else
         hit1_1 = 1'b0;
   end
   
   //
   // Flushing, pipeling and other registered actions
   //
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         flushing    <= 1'b1;
         flushing_1  <= 1'b1;
         flush_count <= {L2MEMDEPTH{1'b0}};
         
         valid_1        <= 1'b0;
         wait_for_write <= 1'b0;
         
         tag_1 <= 'b0;
         
         // Needed to se;ct 'way' update
         mem_valid     <= 1'b0;
         pseudo_random <= 1'b0;
      end
      else      
      begin
         // pipelining
         flushing_1 <= flushing;
         tag_1      <= tag;
         
         if (ready)
            valid_1 <= adrs_valid & !flushing;
         else
            valid_1 <= 1'b0;
            
         // Need to know when we are waiting for a write
         // This is to generate a correct valid_1 signal
         if (valid_1 & !hit0_1 & !hit1_1)
            wait_for_write <= 1'b1;
         else
            if (tag_update)
               wait_for_write <= 1'b0;
         
         // Flush control
         // Multiple flush requests do not start a new flush when
         // the previous one is still in progress.
         // This also means the 'flush' input can be toggled slowly
         // e.g. by a CPU write-set followed by a write-clear
         if (flush & !flushing)
            flushing <= 1'b1;
            // flush counter is zero here
         else
            if (flush_count=={L2MEMDEPTH{1'b1}})
               flushing <= 1'b0;
         
         if (flushing)
            flush_count <= flush_count + 1;
            
         // Need to remember which memories where valid
         if (valid_1)
            mem_valid <= {mem_valid1_1,mem_valid0_1};
            
         // Humm.. could do more but just let it be
         pseudo_random <= ~pseudo_random;
                          
      end
   end
   
   //
   // Decide which of the two 'ways' we write to
   // An empty (not valid) way goes first
   // If both are used select one at (pseudo) random
   //
   always @( * )
   begin
      if (mem_valid[0]==1'b0)
         bank_select = 1'b0;
      else
         if (mem_valid[1]==1'b0)
            bank_select = 1'b1;
         else
            bank_select = pseudo_random;
   end
   
   assign tagmem_write0 = bank_select==1'b0 || flushing ? tagmem_write : 1'b0;
   assign tagmem_write1 = bank_select==1'b1 || flushing ? tagmem_write : 1'b0;
      
endmodule
   
   
