//
// Synchronous tag handling 
// Single bank (1-way) 
// Fast flush version!
// Part of the ro_ff_cache.v design
// 
// Design by G.J. van Loo, FenLogic Ltd, December-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
//
// This module handles the cache tag.
// This is the simplest tag system you can have 
// whilst still using synchronous memories for tag storage.
// It decides if there is a cache hit or miss.
// In case of a miss the tag memory must be updated. 
// This is the fast flush version where we have a sparate 
// 16-bit wide memory for the valid bits 
// This allows us to clear/flush the tag 16 times faster
// 
// Addtional required modules:
//   spm.v (Synchronous singleport memory)
//
  
  
   
module ff_tag
#(parameter
    L2DEPTH      =12,// Log 2 number of cache data entries 
    ADRSWIDTH    =30,// Address bits neede to access cache data units
    L2VALIDWIDTH = 4 // Log 2 valid memory data width 
  )
(
   // System
   input   clk,
   input   reset_n,
   
   // Control
   input       flush,
   output reg  flushing,
   
   // Tag lookup
   input                  adrs_valid,
   input  [ADRSWIDTH-1:0] radrs,
   
   // Match result: one cycle later
   output  reg    valid_1,
   output  reg    hit_1,
   // indicate when we are ready  
   output         ready,
   
   // Tag update 
   input                 tag_update,
   input [ADRSWIDTH-1:0] wt_adrs
   
   );

// For a one-way cache these are identical   
localparam L2MEMDEPTH  = L2DEPTH;

// We need to match all address bits which are not used
// to address the tag memory.
localparam MATCH_WIDTH = ADRSWIDTH - L2DEPTH;

// Valid memory depth and width calculations
localparam L2VALDEPTH   = L2DEPTH-L2VALIDWIDTH;
localparam VALIDWIDTH   = 1<<L2VALIDWIDTH;
// Want to have vector with value 1 but of the right width 
localparam [VALIDWIDTH-1:0] VALID_ONE = { {(VALIDWIDTH-1){1'b0}}, 1'b1}; 

 
// Need address counter plus flag for flushing
reg  [L2VALDEPTH-1:0] flush_count;
reg                   flushing_1;

// Read from tag memory   
reg   [L2MEMDEPTH-1:0] tagmem_adrs;
wire [MATCH_WIDTH-1:0] mem_match_1;
wire                   mem_valid_1;

// Read address for tag memory
wire [L2MEMDEPTH-1:0]  tag_adrs;

// Match-against data now and one cycle later
wire [MATCH_WIDTH-1:0] tag;
reg  [MATCH_WIDTH-1:0] tag_1;

// Write address split into adressa and tag
wire   [L2MEMDEPTH-1:0] write_adrs;
wire  [MATCH_WIDTH-1:0] write_tag;

// Valid memory addresses are smaller
reg    [L2VALDEPTH-1:0] validmem_adrs;
reg  [L2VALIDWIDTH-1:0] valid_rindex_1; // Index valid bit in word
wire   [VALIDWIDTH-1:0] validmem_rdata_1;
wire   [VALIDWIDTH-1:0] validmem_wdata;
wire   [VALIDWIDTH-1:0] valid_wflags;
wire [L2VALIDWIDTH-1:0] valid_windex;  // Index valid bit in word
reg    [VALIDWIDTH-1:0] valid_hold;
wire                    validmem_write;


// The most difficult part: data flow control
// To do with ready & wait_for_write & flushing
reg                    wait_for_write;

   // Split the address in a tag and match
   assign tag_adrs   = radrs[L2MEMDEPTH-1:0];
   assign tag        = radrs[ADRSWIDTH-1:L2MEMDEPTH];
   assign write_adrs = wt_adrs[L2MEMDEPTH-1:0];
   assign write_tag  = wt_adrs[ADRSWIDTH-1:L2MEMDEPTH];
   assign valid_windex = wt_adrs[L2VALIDWIDTH-1:0];
   
   // not ready on miss, waiting or flushing 
   assign ready = ! ((valid_1 & !hit_1) | wait_for_write | flushing);

   // We write the old tag bits but update one:
   assign valid_wflags   = valid_hold | (VALID_ONE << valid_windex);

   // Write to tag memory when flusing or updating
   // All other times we read (Thus no power saving!)   
   assign validmem_write = flushing ? 1'b1 : tag_update;
   assign validmem_wdata = flushing ? {VALIDWIDTH{1'b0}} : valid_wflags;
      
   // multiplex the single ported memory addresses
   // Valid address is a subset (MS bits) 
   always @( * )
   begin
      if (flushing)
      begin
         tagmem_adrs   = tag_adrs;
         validmem_adrs = flush_count;
      end
      else
         if (tag_update)
         begin
            tagmem_adrs = write_adrs;
            validmem_adrs = write_adrs[L2DEPTH-1:L2VALIDWIDTH];
         end
         else
         begin
            tagmem_adrs = tag_adrs;
            validmem_adrs = tag_adrs[L2DEPTH-1:L2VALIDWIDTH];
         end
      
   end


// Synchronous read memory for tags 
spm
  #(
      .WIDTH  (MATCH_WIDTH),    // width in bits
      .L2DEPTH(L2MEMDEPTH)      // Log 2 Depth 
   ) // parameters
tag_mem (
      // write port 
      .clk     (clk),             // write clock  
      .enable  (1'b1),     
               
      .write   (tag_update),     // write access 
      .address (tagmem_adrs),    // read/write address
      .wdata   (write_tag),      // write data
      .rdata   (mem_match_1)     // read data ( 1 cycle delay)
   );

// Synchronous read memory for valid bits 
spm
  #(
      .WIDTH  (VALIDWIDTH),     // width in bits
      .L2DEPTH(L2VALDEPTH)      // Log 2 Depth 
   ) // parameters
valid_mem (
      // write port 
      .clk     (clk),             // write clock  
      .enable  (1'b1),     
               
      .write   (validmem_write),   // write access 
      .address (validmem_adrs),    // read/write address
      .wdata   (validmem_wdata),   // write data
      .rdata   (validmem_rdata_1)  // read data ( 1 cycle delay)
   );

   
   // Pick valid bit from valid word 
   assign mem_valid_1 = validmem_rdata_1[valid_rindex_1]; 
  
   // Here we do the actual match
   // Whilst flushing we cannot have a hit
   always @( * )
   begin
      if (!flushing_1 && mem_valid_1 && tag_1==mem_match_1)
         hit_1 = 1'b1; 
      else
         hit_1 = 1'b0;
   end
   
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         flushing    <= 1'b1;
         flushing_1  <= 1'b1;
         flush_count <= {L2MEMDEPTH{1'b0}};
         
         valid_1        <= 1'b0;
         wait_for_write <= 1'b0;
         
         // Does not need reset
         tag_1 <= 'b0;
         valid_rindex_1 <= {L2VALIDWIDTH{1'b0}};
         valid_hold     <= {VALIDWIDTH{1'b0}};
      end
      else      
      begin
         // pipelining
         flushing_1 <= flushing;
         tag_1      <= tag;
         
         if (ready)
         begin
            valid_1 <= adrs_valid & !flushing;
            valid_rindex_1 <= tag_adrs[L2VALIDWIDTH-1:0];
         end
         else
            valid_1 <= 1'b0;
            
        // Hold the valid word on miss as we have to write it again 
        if (valid_1 & !hit_1)
            valid_hold <= validmem_rdata_1;
                    
         // Need to know when we are waiting for a write
         // This is to generate a correct valid_1 signal
         if (valid_1 & !hit_1)
            wait_for_write <= 1'b1;
         else
            if (tag_update)
               wait_for_write <= 1'b0;
         
         // Flush control
         // Multiple flush requests do not start a new flush when
         // the previous one is still in progress.
         // This also means the 'flush' input can be toggled slowly
         // e.g. by a CPU write-set followed by a write-clear
         if (flush & !flushing)
            flushing <= 1'b1;
            // flush counter is zero here
         else
            if (flush_count=={L2VALDEPTH{1'b1}})
               flushing <= 1'b0;
         
         if (flushing)
            flush_count <= flush_count + 1;
                          
      end
   end
   
endmodule
   
   
