//
// Testbench for asyncheronoous read only cache
// 
// Design by G.J. van Loo, FenLogic Ltd, December-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
// 

module async_ro_cache_test;

localparam CLK_PERIOD=100;

localparam DATAWIDTH= 32; // Width of data
localparam ADRSWIDTH= 16; // Width of address
localparam L2DEPTH  = 8; // log-2 size of the cache 

// start test with static main memory response time
localparam MEM_READ_DELAY = 3; // clock cycles


   // System
reg                  clk;
reg                  reset_n;

// control
reg                 flush;
   
   
   // Read port
reg                  read;
reg  [ADRSWIDTH-1:0] radrs;
wire                 valid;
wire [DATAWIDTH-1:0] rdata;
   
   // External (main) memory port
wire                 mem_read;
wire [ADRSWIDTH-1:0] mem_radrs;
reg                  mem_data_valid;
wire [DATAWIDTH-1:0] mem_rdata;



// testbench variables

integer r;
// Assuming DATAWIDTH > ADRSWIDTH here! 
wire [DATAWIDTH-1:0]  long_adrs;
wire    real_hit;
integer f,f_count;

   initial
   begin

   // System
      clk       = 1'b0;
      reset_n   = 1'b0;
      
   // control
      flush     = 1'b0;
   
   
   // Read port
      read            = 1'b0;
      radrs           = 'b0;
      #(5*CLK_PERIOD) reset_n=1'b1;
      // Fill main memory with address==data
      for (r=0; r<(1<<ADRSWIDTH); r=r+1)
        main_mem.do_write(r,r);

      // Read address 0-15 then again
      // gives 16 misses followed by 16 hits
      // One dead cycle between reads. 
      #(5*CLK_PERIOD);
      for (r=0; r<32; r=r+1)      
      begin
         radrs <= r%16; 
         @(posedge clk) ; 
         read <= 1'b1;
         @(posedge clk) ;
         while (valid==1'b0)
            @(posedge clk);     
         read <= 1'b0;
      end
      #(10*CLK_PERIOD);
      
      // Read address 0-32 then again
      // gives 16 hits, 16 misses then 32 hits
      // No dead cycle between reads. 
      #(5*CLK_PERIOD);
      for (r=0; r<64; r=r+1)      
      begin
         radrs <= r%32; 
         @(posedge clk) ; 
         read <= 1'b1;
         while (valid==1'b0)
            @(posedge clk);     
      end
      #(10*CLK_PERIOD);
     
      // Random reads
      // No dead cycle between reads. 
      #(5*CLK_PERIOD);
      for (r=0; r<1000000; r=r+1)      
      begin
         // Limit address to get more hits
         radrs <= $urandom & 32'h000003FF;
         @(posedge clk) ; 
         #1; 
         read <= 1'b1;
         while (valid==1'b0)
            @(posedge clk);     
      end
      #(10*CLK_PERIOD);
      
      // Flush test
      #(5*CLK_PERIOD);
      for (f=0; f<100; f=f+1)
      begin
         f_count = $urandom & 32'h000000FF;
         for (r=0; r<5000; r=r+1)      
         begin
            f_count = f_count - 1;
            if (f_count==0)
               flush <= 1'b1;
            // Limit address to get more hits
            radrs <= $urandom & 32'h000003FF;
            @(posedge clk) ; 
            #1; 
            read <= 1'b1;
            while (valid==1'b0)
            begin
               f_count = f_count - 1;
               if (f_count==0)
                  flush <= 1'b1;
               if (flush==1'b1)
                  flush <= 1'b0;
               @(posedge clk);     
            end
         end
         #(10*CLK_PERIOD);
      end
      
    
      // If a test fails it stops at that point
      // Thus if we arrive here we have found no errors.
      // (Which is different from: there are no errors!!)
      $display("************************");
      $display("**                    **");
      $display("**    Test Passed     **");
      $display("**                    **");
      $display("************************");
      $stop;
   end


async_ro_cache

   #(
      .DATAWIDTH (DATAWIDTH),  // Width of data
      .ADRSWIDTH (ADRSWIDTH),  // Width of address
      .L2DEPTH   (L2DEPTH)     // log-2 size of the cache 

   ) // parameters
async_ro_cache_0 (

   // System
      .clk           (clk),
      .reset_n       (reset_n),
      
      .flush         (flush),
   
   
   // Read port
      .read          (read),
      .radrs         (radrs),
      .valid         (valid),
      .rdata         (rdata),
   
   // External (main) memory port
      .mem_read      (mem_read),
      .mem_radrs     (mem_radrs),
      .mem_data_valid(mem_data_valid),
      .mem_rdata     (mem_rdata) 
   );


//
// This is the main memory 
// 

spm
   #(
      .WIDTH   (DATAWIDTH),  // width in bits
      .L2DEPTH (ADRSWIDTH),  // Log 2 Depth, 8 = 256 deep 
      // Simulation parameters
      .CLEAR (0),
      .FILE  ("") 
   ) // parameters
main_mem (
      .clk    (clk),     // system clock 
               
      .enable (mem_read),  // access 
      .write  (1'b0),      // write 
      .address(mem_radrs), // read/write address
      .wdata  (32'h0),     // write data
      .rdata  (mem_rdata) 
   );


integer delay;

   // 
   // Make memory respond with delay 
   //
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         mem_data_valid <= 1'b0;
         delay          <= 0;
      end
      else
      begin
         if (mem_data_valid)
            mem_data_valid <= 1'b0;
         else
            if (delay==1)
               mem_data_valid <= 1'b1;
               
         if (delay>=0)
            delay <= delay - 1;
         else
            if (mem_read)
               delay <= MEM_READ_DELAY+1;
      end
   end

   // Check results
   assign long_adrs = radrs; // expand address to 32 bits
   // cache will be empty giving nothing but misses
   always @(posedge clk)
   begin
      if (read & valid && long_adrs!==rdata)
      begin
         $display("Cache error @%0t. Have 0x%08x, expected 0x%08x",
                    $time,rdata,long_adrs);
         #1 $stop; 
      end
   end
   

/*   
// Estimate hit ratio
integer wait_a_while;
integer accesses,hits;
real    ra,rh,ratio; 

   // wait a while before starting as
   // cache will be empty giving nothing but misses
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         wait_a_while <= 1024;
         accesses <= 0;
         hits     <= 0;
         
      end
      else
      begin
         if (wait_a_while!=0)
            wait_a_while = wait_a_while -1;
         if (wait_a_while==0)
         begin
            if (read & rdata_valid)
               accesses = accesses + 1;
               
            if (read & real_hit)
               hits <= hits + 1;
         end
      end
   end
*/
 
   // Generate clock.
   initial
   begin
      clk = 1'b0;
      forever
         #(CLK_PERIOD/2) clk = ~clk;
   end

endmodule
