//
// Testbench for read only cache
// 
// Design by G.J. van Loo, FenLogic Ltd, November-2017.
// Mail: verilog@fenlogic.com
//
// This program is free software. You can use and redistribute 
// the program and or modify it in any form without ANY obligations.
// But the author would appreciated if the credits stays in.
// As ususal, it comes without any guarantees or warranty to the extent 
// permitted by applicable law. Although the author has attempted to find
// and correct any bugs in this free software program, the author is not
// responsible for any damage or losses of any kind caused by the use 
// or misuse of the program. 
// 

module ro_cache_test;

localparam CLK_PERIOD=100;

localparam DATAWIDTH= 32; // Width of data
localparam ADRSWIDTH= 16; // Width of address
localparam L2DEPTH  = 8;  // Log 2 number of cache data entries 

// start test with static main memory response time
localparam MEM_READ_DELAY = 3; // clock cycles


   // System
reg                  clk;
reg                  reset_n;

// control
reg                 flush;
   
   
   // Read port
reg                  read;
reg  [ADRSWIDTH-1:0] radrs;
wire                 ready;
wire [DATAWIDTH-1:0] rdata;
   
   // External (main) memory port
wire                 mem_read;
wire [ADRSWIDTH-1:0] mem_radrs;
reg                  mem_data_valid;
wire [DATAWIDTH-1:0] mem_rdata;
wire [DATAWIDTH-1:0] valid_mem_rdata;



// testbench variables

integer r;
// Assuming DATAWIDTH > ADRSWIDTH here! 
wire [DATAWIDTH-1:0] long_adrs;
reg  [ADRSWIDTH-1:0] address_mask;

// Pick up result one cycle after emitted (if ready)
reg                   expect;
reg  [ADRSWIDTH-1:0]  radrs_1;

// Statistics
integer wait_a_while;
integer accesses,misses;
reg     read_1;
reg     stats_on;
real    ra,rh,ratio; 

// Special sequences
integer adrs_index;

// flush test
integer f,f_count;


   initial
   begin
   // testbench 
      stats_on  = 1'b0;

   // System
      clk       = 1'b0;
      reset_n   = 1'b0;
      
   // control
      flush     = 1'b0;
   
   
   // Read port
      read            = 1'b0;
      radrs           = 'b0;
      #(5*CLK_PERIOD) reset_n=1'b1;
      // Fill main memory with address==data
      for (r=0; r<(1<<ADRSWIDTH); r=r+1)
        main_mem.do_write(r,r);

      // Read address 0-15 then again
      // gives 16 misses followed by 16 hits
      // One dead cycle between reads. 
      #(5*CLK_PERIOD);
      wait (ready===1'b1);
      #(5*CLK_PERIOD);
      
      @(posedge clk) ;          
      for (r=0; r<32; r=r+1)      
      begin
         radrs <= r%16; 
         read <= 1'b1;
         @(posedge clk) ;          
         radrs <= 'hx; // so we don't accidentally use it! 
         read  <= 1'b0;
         @(posedge clk) ;
         while (ready==1'b0)
            @(posedge clk);     
      end
      read  <= 1'b0;
      #(10*CLK_PERIOD);
      
      // Read address 0-32 then again
      // gives 16 hits, 16 misses then 32 hits
      // No dead cycle between reads.
      // ============================ 
      
      // Output first address unconditionally 
      @(posedge clk) ; 
      radrs <= 0; 
      read  <= 1'b1;
      @(posedge clk) ; 
      // Next ones are output and waiting for ready
      for (r=1; r<64; r=r+1)      
      begin
         radrs <= r%32; 
         read  <= 1'b1;
         @(posedge clk) ; 
         while (ready==1'b0)
            @(posedge clk);     
      end
      read <= 1'b0;
      #(10*CLK_PERIOD);
      

      //
      // special sequence reads
      //
      @(posedge clk) ; 
      radrs <= 0; 
      read  <= 1'b1;
      @(posedge clk) ; 
      for (adrs_index=0; adrs_index<34; adrs_index=adrs_index+1)    
      begin
         radrs <= address_rom(adrs_index); 
         read  <= 1'b1;
         @(posedge clk) ; 
         while (ready==1'b0)
            @(posedge clk);     
      end
      read <= 1'b0;
      #(10*CLK_PERIOD);
      
//      $stop;
      
      // Random reads
      // No dead cycle between reads. 
      stats_on = 1;
      address_mask = (1<<(L2DEPTH+2))-1; // 25% hit probability:
      #(5*CLK_PERIOD);
      for (r=0; r<1000000; r=r+1)      
      begin
         // Limit address to get more hits
         radrs <= $urandom & address_mask;
         read <= 1'b1;
         @(posedge clk) ; 
         while (ready==1'b0)
            @(posedge clk);     
      end
      read <= 1'b0;
      #(10*CLK_PERIOD);
      stats_on = 0;

      // Flush test
      #(5*CLK_PERIOD);
      for (f=0; f<100; f=f+1)
      begin
         f_count = $urandom & 32'h000000FF;
         for (r=0; r<5000; r=r+1)      
         begin
            f_count = f_count - 1;
            if (f_count==0)
               flush <= 1'b1;
            // Limit address to get more hits
            radrs <= $urandom & address_mask;
            read <= 1'b1;
            @(posedge clk) ; 
            while (ready==1'b0)
            begin
               f_count = f_count - 1;
               if (f_count==0)
                  flush <= 1'b1;
               if (flush==1'b1)
                  flush <= 1'b0;
               @(posedge clk);     
            end
         end
         #(10*CLK_PERIOD);
      end

      // If a test fails it stops at that point
      // Thus if we arrive here we have found no errors.
      // (Which is different from: there are no errors!!)
      ra = accesses;
      rh = accesses - misses;
      ratio = rh/ra;
      $display("Reads=%d  Hits=%d   Ratio=%f",accesses,accesses-misses,ratio);
      $display("************************");
      $display("**                    **");
      $display("**    Test Passed     **");
      $display("**                    **");
      $display("************************");
      $stop;

   end


ro_cache

   #(
      .DATAWIDTH (DATAWIDTH),  // Width of data
      .ADRSWIDTH (ADRSWIDTH),  // Width of address
      .L2DEPTH   (L2DEPTH)     // log-2 size of the cache 

   ) // parameters
ro_cache_0 (

   // System
      .clk           (clk),
      .reset_n       (reset_n),
      
      .flush         (flush),
   
   
   // Read port
      .read          (read),
      .radrs         (radrs),
      .ready         (ready),
      .rdata         (rdata),
   
   // External (main) memory port
      .mem_read      (mem_read),
      .mem_radrs     (mem_radrs),
      .mem_data_valid(mem_data_valid),
      .mem_rdata     (valid_mem_rdata) 
   );


//
// This is the main memory 
// 

spm
   #(
      .WIDTH   (DATAWIDTH),  // width in bits
      .L2DEPTH (ADRSWIDTH),  // Log 2 Depth, 8 = 256 deep 
      // Simulation parameters
      .CLEAR (0),
      .FILE  ("") 
   ) // parameters
main_mem (
      .clk    (clk),     // system clock 
               
      .enable (mem_read),  // access 
      .write  (1'b0),      // write 
      .address(mem_radrs), // read/write address
      .wdata  (32'h0),     // write data
      .rdata  (mem_rdata) 
   );
   
   // make data X if valid not presetn
   assign valid_mem_rdata = mem_data_valid ? mem_rdata : 'hx;


integer delay;

   // 
   // Make memory respond with delay 
   //
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         mem_data_valid <= 1'b0;
         delay          <= 0;
      end
      else
      begin
         if (mem_data_valid)
            mem_data_valid <= 1'b0;
         else
            if (delay==1)
               mem_data_valid <= 1'b1;
               
         if (delay>=0)
            delay <= delay - 1;
         else
            if (mem_read)
               delay <= MEM_READ_DELAY+1;
      end
   end

   // 
   // Special sequence of addresses to get
   // specific hit/miss patterns. 
   //
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         mem_data_valid <= 1'b0;
         delay          <= 0;
      end
      else
      begin
         if (mem_data_valid)
            mem_data_valid <= 1'b0;
         else
            if (delay==1)
               mem_data_valid <= 1'b1;
               
         if (delay>=0)
            delay <= delay - 1;
         else
            if (mem_read)
               delay <= MEM_READ_DELAY+1;
      end
   end


   //
   // Check if result is correct
   // Because address==data this is easy.
   //
   assign long_adrs = radrs_1; // expand address to 32 bits
   // cache will be empty giving nothing but misses
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         radrs_1 <= 'hx;
         expect  <= 1'b0;
      end
      else
      begin
         
         if (ready)
         begin
            radrs_1  <= radrs;
            expect   <= read;
         end
            
            
         if (expect && ready && long_adrs!==rdata)
         begin
            $display("Cache error @%0t. Have 0x%08x, expected 0x%08x",
                       $time,rdata,long_adrs);
            #1 $stop; 
         end
      end
   end
   

   //
   // Check statistics of result to compare
   // against expectations
   //

   // wait a while before starting as
   // cache will be empty giving nothing but misses
   // that distorts the statistics
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         wait_a_while <= 6000; // clocks, not reads! 
         accesses <= 0;
         misses   <= 0;
         read_1   <= 1'b0;
      end
      else
      begin
         
         if (ready)
            read_1 <= read;
            
         if (wait_a_while!=0)
            wait_a_while = wait_a_while -1;
         if (wait_a_while==0)
         begin
            if (read_1 & ready)
               accesses = accesses + 1;
               
            if (mem_read)
               misses <= misses+ 1;
               
         end
         
         // This overrides all:
         if (stats_on==0)
            wait_a_while <= 6000;
         
      end
   end
 
   // Generate clock.
   initial
   begin
      clk = 1'b0;
      forever
         #(CLK_PERIOD/2) clk = ~clk;
   end

//
// Address ROM for generating special test sequences
// This one works only:
// - If the memory has not be access from 256 upwards
// - The cache depth is 256
// This is to test address sequences which are expected
// to be troublesome (e.g. same tag address but different tags)
// 
function [ADRSWIDTH-1:0] address_rom(input integer index);
begin
   case (index)
     0 : address_rom = 256; // miss
     1 : address_rom = 256; // hit following previous miss
     2 : address_rom = 257; // miss (giving miss, hit, miss sequence)
     3 : address_rom = 257; // hit  (giving miss, hit, miss, hit sequence)
    // new sequence: repeated use of same tag address but different tag
    // misses in between 
     4 : address_rom = 270; // miss
     5 : address_rom = 271; // miss
     6 : address_rom = 272; // miss
     7 : address_rom = 270+256; // miss but using same tag address as 270 miss!
    // again with one less miss in the middle
     8 : address_rom = 280; // miss 
     9 : address_rom = 281; // miss
    10 : address_rom = 280+256; // miss but using same tag address as 280 miss!
    // again with no miss in the middle
    11 : address_rom = 290; // miss 
    12 : address_rom = 290+256; // miss but using same tag address as 290 miss!
        // new sequence: repeated use of same tag address but different tag
        // misses in between 
    13 : address_rom = 270; // miss
    14 : address_rom = 270; // hit
    15 : address_rom = 291; // miss
    16 : address_rom = 292; // miss
    17 : address_rom = 270+256; // miss but using same tag address as 270 hit!
    // again with one less miss in the middle
    18 : address_rom = 280; // miss 
    19 : address_rom = 280; // hit 
    20 : address_rom = 293; // miss
    21 : address_rom = 280+256; // miss but using same tag address as 280 hit!
    // again with no miss in the middle
    22 : address_rom = 290; // miss 
    23 : address_rom = 290; // hit 
    24 : address_rom = 290+256; // miss but using same tag address as 290 hit!
        // new sequence: repeated use of same tag address but different tag
        // hits in between 
    25 : address_rom = 270; // miss
    26 : address_rom = 270; // hit
    27 : address_rom = 291; // hit
    28 : address_rom = 292; // hit
    29 : address_rom = 270+256; // miss but using same tag address as 270 hit!
    // again with one less miss in the middle
    30 : address_rom = 280; // miss 
    31 : address_rom = 280; // hit 
    32 : address_rom = 291; // hit
    33 : address_rom = 280+256; // miss but using same tag address as 280 hit!
    default : address_rom = {ADRSWIDTH{1'bx}};
   endcase
end
endfunction


endmodule
